local folder, core = ...

local _G = _G
local pairs_ = pairs
local GetTime_ = GetTime
local CreateFrame_ = CreateFrame
local table_remove = table.remove
local table_sort = table.sort
local type = type
local table_getn = table.getn
local Debug = core.Debug

-- Isolate the environment

--~ setfenv(1, core)
-- local

core.buffBars = {}


local P = {}
local buffFrames = core.buffFrames
local LibNameplate = core.LibNameplate
local nametoGUIDs = core.nametoGUIDs
local buffBars = core.buffBars
local buffFrames = core.buffFrames
local guidBuffs = core.guidBuffs

local prev_OnEnable = core.OnEnable;
function core:OnEnable()
	prev_OnEnable(self);
	
--~ 	Debug("OnEnable", "frames", self.db.profile)
	P = self.db.profile --this can change on profile change.
end

local function iconOnShow(self)
--~ 	self.cooldown:SetText("0")--blank
--~ 	Debug("iconOnShow", self.cooldown:GetStringHeight())
	self.alphaUp = false
	self:SetAlpha(1)
	if P.showCooldown == true and self.expirationTime > 0 then
		self.cdbg:Show()
		self.cd:Show()
	else
		self.cdbg:Hide()
		self.cd:Hide()
	end
end


local function iconOnUpdate(self)
	self.lastUpdate = self.lastUpdate + 1
	if self.lastUpdate > 1 then
		self.lastUpdate = 0

		if self.expirationTime > 0 then
			local timeLeft = core:Round(self.expirationTime - GetTime_())
			if P.showCooldown == true then
				self.cd:SetText(core:SecondsToString(timeLeft, 1))
				self.cd:SetTextColor(core:RedToGreen(timeLeft, self.duration))
				self.cdbg:SetWidth(self.cd:GetStringWidth())
			end
			
			if (timeLeft / self.duration ) < .2 then --buff only has 20% timeleft.
				local alpha = self:GetAlpha()
				if alpha <= 0.2 then
					self.alphaUp = true
				elseif alpha >= 1 then
					self.alphaUp = false
				end
				if self.alphaUp == true then
					self:SetAlpha(alpha+0.1)
				else
					self:SetAlpha(alpha-0.1)
				end
			end
			
			if GetTime_() > self.expirationTime then
				self:Hide()
				
				local GUID = LibNameplate:GetGUID(self.realPlate)
				if GUID then
					core:AddBuffsToPlate(self.realPlate, GUID)
				else
					local plateName =  LibNameplate:GetName(self.realPlate)
					if plateName and nametoGUIDs[plateName] then
						core:AddBuffsToPlate(self.realPlate, nametoGUIDs[plateName])
					end
				end
			end
--~ 			else
--~ 				self.cd:Hide()
--~ 				self.cdbg:Hide()
		end
		
	end
end

local function UpdateBuffSize(frame, size)
	frame.texture:SetWidth(size)
	frame.texture:SetHeight(size)
	
	frame:SetWidth(size)
	
	if P.showCooldown == true then
		frame:SetHeight(size + frame.cd:GetStringHeight())
	else
		frame:SetHeight(size)
	end
end

local iX = 0 --used for testing.

function core:GetCooldownFont()
	if type(P.cooldownSize) == "string" then
		P.cooldownSize = 1
	end
	local fonts = {
		"SystemFont_Tiny",
		"SystemFont_Small",
		"SystemFont_Med1",
		"SystemFont_Med2",
		"SystemFont_Med3",
		"SystemFont_Large",
		"SystemFont_Huge1",

	}
	
--~ 	Debug("GetCooldownFont", P.cooldownSize, fonts[P.cooldownSize])
	
	return _G[fonts[P.cooldownSize]]
	
--~ 			["SystemFont_Tiny"] = "Tiny", --9
--~ 			["SystemFont_Small"] = "Small", --10
--~ 			["SystemFont_Med1"] = "Medium", --12
--~ 			["SystemFont_Med2"] = "", --13
--~ 			["SystemFont_Med3"] = "Medium2", --14
--~ 			["SystemFont_Large"] = "Large", --16
--~ 			["SystemFont_Huge1"] = "Huge", --20
end

local function CreateBuffFrame(parentFrame, realPlate)
	local f = CreateFrame_("frame", "PB"..iX, parentFrame) --
	f.realPlate = realPlate
	f:SetFrameStrata("LOW")
	
	f.texture = f:CreateTexture(nil,"BACKGROUND")

	f.texture:SetPoint("TOP", f)
	
	
	local cd = f:CreateFontString(nil, "ARTWORK")
--~ 	cd:SetFontObject(_G[P.cooldownSize])
	cd:SetFontObject(core:GetCooldownFont())
	

	
	cd:SetText("0")
	cd:SetPoint("TOP", f.texture, "BOTTOM")
	cd:Show()
	cd:Hide()
--~ 	cd:SetTextColor(0,1,0,1)
	f.cd = cd

	
	
	--Make the text easier to see.
	f.cdbg = f:CreateTexture(nil,"LOW")
	f.cdbg:SetTexture(0,0,0,.75)
	f.cdbg:SetPoint("CENTER", cd)
	f.cdbg:SetHeight(cd:GetStringHeight())
	f.cdbg:Hide()
	

	UpdateBuffSize(f, P.iconSize)
	
	
	
--~ 	Debug("CreateBuffFrame", f:GetWidth(), f:GetHeight())
	
	f.lastUpdate = 0
	f.expirationTime = 0
	f:SetScript("OnShow", iconOnShow)
	f:SetScript("OnUpdate", iconOnUpdate) 
	
	return f
end

function core:UpdateBarsBackground()
	for plate in pairs_(buffBars) do 
		for b in pairs_(buffBars[plate]) do 
		
			if core.showBarBackground == true then
				buffBars[plate][b].barBG:Show()
			else
				buffBars[plate][b].barBG:Hide()
			end
		
		end
	end
end

local function CreatePlateBar(parentFrame, realPlate)
	iX = iX + 1
	local f = CreateFrame_("frame", "PB"..iX, parentFrame) --
	f.realPlate = realPlate
	
	f:SetFrameStrata("BACKGROUND")
	f:SetFrameLevel(parentFrame:GetFrameLevel())
	
	f:SetWidth(1)
	f:SetHeight(1)
	
	--Make the text easier to see.
	f.barBG = f:CreateTexture(nil,"LOW")
	f.barBG:SetAllPoints(true)

	f.barBG:SetTexture(1,1,1,0.3) --uses for testing
	if core.showBarBackground == true then
		f.barBG:Show()
	else
		f.barBG:Hide()
	end
	
	f:Show()
	return f
end

local function BuildPlateBars(plate, visibleFrame)
	buffBars[plate] = buffBars[plate] or {}
	
	
	if not buffBars[plate][1] then
		buffBars[plate][1] = CreatePlateBar(visibleFrame, plate)
	end
	buffBars[plate][1]:ClearAllPoints()
	buffBars[plate][1]:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY) -- + GetCooldownHeight(buffFrames[plate][1])
	
	
	if P.numRows > 1 then
		for r=2, P.numRows do
			if not buffBars[plate][r] then
				buffBars[plate][r] = CreatePlateBar(visibleFrame, plate)
			end
			buffBars[plate][r]:ClearAllPoints()
			
			
			if P.rowGroth == 1 then --up
--~ 				buffBars[plate][r]:SetPoint("BOTTOMLEFT", buffFrames[plate][total - P.iconsPerRow], "TOPLEFT", 0, GetCooldownHeight(buffFrames[plate][total])) --
				
				buffBars[plate][r]:SetPoint("BOTTOMLEFT", buffBars[plate][r-1], "TOPLEFT", 0, 0)
			else --down
--~ 				buffBars[plate][r]:SetPoint("TOPLEFT", buffFrames[plate][total - P.iconsPerRow], "BOTTOMLEFT", 0, GetCooldownHeight(buffFrames[plate][total]))
				buffBars[plate][r]:SetPoint("TOPLEFT", buffBars[plate][r-1], "BOTTOMLEFT", 0, 0)
			end
		end
	end
end

local function GetTimeleft(expirationTime)
	return core:Round(expirationTime - GetTime_())


end

local highNumber = 9999999999999999999

-- Set textures and show the frames.
function core:AddBuffsToPlate(plate, GUID)
	if not buffFrames[plate] or not buffFrames[plate][2]  then
		self:BuildBuffFrame(plate)
	end
	
	local t, f

--~ 	Debug("AddBuffsToPlate 0", GUID)
	if guidBuffs[GUID] then
	
		for i=(P.numRows * P.iconsPerRow), 1, -1 do 
			if guidBuffs[GUID][i] then
				if guidBuffs[GUID][i].expirationTime > 0 and GetTime_() > guidBuffs[GUID][i].expirationTime then
					table_remove(guidBuffs[GUID], i)
				end
			end
		end
	
		table_sort(guidBuffs[GUID], function(a,b) 
			if(a and b) then 
				return (a.playerCast or 0) > (b.playerCast or 0);
			end 
		end)

	
--~ 		Debug("AddBuffsToPlate 1", GUID)
		for i=1, P.numRows * P.iconsPerRow do 
--~ 			Debug("AddBuffsToPlate 2", GUID)
			if buffFrames[plate][i] then
				if guidBuffs[GUID][i] and (guidBuffs[GUID][i].expirationTime == 0 or guidBuffs[GUID][i].expirationTime > GetTime_()) then
	--~ 				Debug("AddBuffsToPlate 3", i, GUID)
					buffFrames[plate][i].texture:SetTexture(guidBuffs[GUID][i].icon)
	
					buffFrames[plate][i].expirationTime = guidBuffs[GUID][i].expirationTime or 0
					buffFrames[plate][i].duration = guidBuffs[GUID][i].duration or 1
					
					if guidBuffs[GUID][i].playerCast and P.biggerSelfSpells == true then
--~ 						buffFrames[plate][i].texture:SetWidth(P.iconSize * 1.2)
--~ 						buffFrames[plate][i].texture:SetHeight(P.iconSize * 1.2)
						UpdateBuffSize(buffFrames[plate][i], P.iconSize * 1.2)
					else
--~ 						buffFrames[plate][i].texture:SetWidth(P.iconSize)
--~ 						buffFrames[plate][i].texture:SetHeight(P.iconSize)
						UpdateBuffSize(buffFrames[plate][i], P.iconSize)
					end
					
					if buffFrames[plate][i]:IsShown() then
						iconOnShow(buffFrames[plate][i])
					else
						buffFrames[plate][i]:Show()--make sure OnShow fires.
					end
				else
	--~ 				Debug("AddBuffsToPlate 4", i, GUID)
					buffFrames[plate][i]:Hide()
					buffFrames[plate][i].expirationTime = 0
				end
			end
		end
	end
end

--display question mark icon
function core:AddUnknownIcon(plate)
	if not buffFrames[plate] then
		self:BuildBuffFrame(plate, nil, true)
	end
	
	buffFrames[plate][1].texture:SetTexture("Interface\\Icons\\Inv_misc_questionmark")
	buffFrames[plate][1].expirationTime = 0

	if buffFrames[plate][1]:IsShown() then
		iconOnShow(buffFrames[plate][1])
	else
		buffFrames[plate][1]:Show()--make sure OnShow fires.
	end
end

function core:ResetAllPlateIcons()
	for plate in pairs_(buffFrames) do 
		core:BuildBuffFrame(plate, true)
	end
end


function core:BuildBuffFrame(plate, reset, onlyOne)
	local visibleFrame = LibNameplate:GetVisibleFrame(plate)
	if not buffBars[plate] then
		BuildPlateBars(plate, visibleFrame)
	end

	if not buffBars[plate][P.numRows] then--user increased the size.
		BuildPlateBars(plate, visibleFrame)
	end
	
	
	buffFrames[plate] = buffFrames[plate] or {}
	
	if reset then
		for i=1, table_getn(buffFrames[plate]) do 
			buffFrames[plate][i]:Hide() --makesure all frames stop OnUpdating.
			
--~ 			buffFrames[plate][i]:SetParent(nil)
		end
		
		for b in pairs_(buffBars[plate]) do 
			buffBars[plate][b]:Hide()
		end
	end
	
	local total = 1 --total number of spell frames
	if not buffFrames[plate][total] then
		buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][1], plate)
	end
	buffFrames[plate][total]:ClearAllPoints()
	buffFrames[plate][total]:SetPoint("TOPLEFT", buffBars[plate][1]) -- + GetCooldownHeight(buffFrames[plate][1])
	buffBars[plate][total]:Show()
	
	if onlyOne then
		self:UpdateBarsSize()
		return
	end

	
	local prevFrame = buffFrames[plate][total]
	for i=2, P.iconsPerRow do 
		total = total + 1
		if not buffFrames[plate][total] then
			buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][1], plate)
		end
		buffFrames[plate][total]:SetParent(buffBars[plate][1])

		buffFrames[plate][total]:ClearAllPoints()
		buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")

		prevFrame = buffFrames[plate][total]
	end
	
	if P.numRows > 1 then
		for r=2, P.numRows do 

			for i=1, P.iconsPerRow do 
				total = total + 1

				if not buffFrames[plate][total] then
					buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][r], plate)
				end
				buffFrames[plate][total]:SetParent(buffBars[plate][r])
				
				
				buffFrames[plate][total]:ClearAllPoints()
				if i == 1 then
					
--~ 					if P.rowGroth == 1 then --up
						buffFrames[plate][total]:SetPoint("TOPLEFT", buffBars[plate][r])
--~ 					else
--~ 						buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")
--~ 					end
					
					
					
				else
					buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")
				end
				
				
				
				prevFrame = buffFrames[plate][total]
				
				
				
			end
			buffBars[plate][r]:Show()
		
		end
	end
	
	self:UpdateBarsSize()
end

function core:ResetBarPoint(barFrame, plate, visibleFrame)
	barFrame:ClearAllPoints()
--~ 			buffFrames[plate][i]:SetPoint(P.barAnchorPoint, plate, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY)
	local visibleFrame = visibleFrame or LibNameplate:GetVisibleFrame(plate)
	barFrame:SetParent(visibleFrame)
	barFrame:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY) -- + GetCooldownHeight(barFrame)

--~ 	barFrame:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY)
end

local table_getn = table.getn
local function UpdateBarSize(barFrame)
--~ 	local c = {barFrame:GetChildren()}
	
	local eachWidth = 0
	local eachHeight = 0
	
	for plate in pairs_(buffFrames) do 
		if buffFrames[plate][1] then
			eachWidth	= buffFrames[plate][1]:GetWidth()
			eachHeight	= buffFrames[plate][1]:GetHeight()

			break
		end
	end
	--***
	
	
	
	
	
	
--~ 	local totalWidth = 0
--~ 	for i = 1, table_getn(c) do
--~ 		totalWidth = totalWidth + c[i]:GetWidth()
--~ 	end
	
	
	
	barFrame:SetHeight(eachHeight)
	barFrame:SetWidth(eachWidth * P.iconsPerRow)
	
--~ 	Debug("UpdateBarSize", barFrame:GetHeight(), barFrame:GetWidth())
end

function core:UpdateBarsSize()
	for plate in pairs_(buffFrames) do 
		for i=1, table_getn(buffBars[plate]) do 
			UpdateBarSize(buffBars[plate][i])
		end
	end
end

function core:ResetIconSizes()
	local size = P.iconSize
	local frame
	for plate in pairs_(self.buffFrames) do 
		for i=1, table_getn(self.buffFrames[plate]) do 
			frame = self.buffFrames[plate][i]
			
			--Update the spell texture size.
			frame.texture:SetWidth(size)
			frame.texture:SetHeight(size)

			--Update the frame as a whole, this takes into account the size of the cooldown size.
			frame:SetWidth(size)
			
			if P.showCooldown == true then
				frame:SetHeight(size + frame.cd:GetStringHeight())
			else
				frame:SetHeight(size)
			end
		end
	end
end

function core:ResetCooldownSize()
	for plate in pairs_(buffFrames) do 
		for i=1, table_getn(buffFrames[plate]) do 
			buffFrames[plate][i].cd:SetFontObject(core:GetCooldownFont())
			buffFrames[plate][i].cdbg:SetHeight(buffFrames[plate][i].cd:GetStringHeight())
			iconOnShow(buffFrames[plate][i]) --lazy way to hide/show cooldown.
		end
	end
end

function core:ResetAllBarPoints()
	for plate in pairs_(buffBars) do 
		if buffBars[plate][1] then
			self:ResetBarPoint(buffBars[plate][1], plate)
		end
		
		for r=2, table_getn(buffBars[plate]) do 
			buffBars[plate][r]:ClearAllPoints()
			if P.rowGroth == 1 then --up
				buffBars[plate][r]:SetPoint("BOTTOMLEFT", buffBars[plate][r-1], "TOPLEFT", 0, 0)
			else --down
				buffBars[plate][r]:SetPoint("TOPLEFT", buffBars[plate][r-1], "BOTTOMLEFT", 0, 0)
			end
		end

		
		-- ****
	end
--~ 	for plate in pairs_(buffFrames) do 
--~ 		if buffFrames[plate][1] then
--~ 			self:ResetBarPoint(buffFrames[plate][1], plate)
--~ 		end
--~ 	end
end

function core:ShowAllKnownSpells()
-- When we change number of icons to show we hide all icons. This will reshow the buffs again in their new locations.
	local GUID
	for plate in pairs_(buffFrames) do 
		local GUID = LibNameplate:GetGUID(plate)
		if GUID then
			self:AddBuffsToPlate(plate, GUID)
		else
			local plateName =  LibNameplate:GetName(plate)
			if plateName and nametoGUIDs[plateName] then
				self:AddBuffsToPlate(plate, nametoGUIDs[plateName])
			end
		end
	end
end